#---------------------------------------------------------------------
# authentication
#---------------------------------------------------------------------

package AbsenceAuthentication;

use FileHandle;
use Fcntl ':flock'; # import LOCK_* constants
use Carp;

BEGIN {
	# an instance has been specified
	if (exists($ENV{INSTANCE_NAME})) {
        $COOKIE_NAME = "absence-auth-$ENV{INSTANCE_NAME}";
    } else {
        $COOKIE_NAME = 'absence-auth';
    }
}

use AbsenceConfig;
use AbsenceDB;
use AbsenceLog;


use Digest::MD5 qw(md5_base64);

#---------------------------------------------------------------------
# package globals
#---------------------------------------------------------------------

my $AUTH_TYPE		= AbsenceConfig::fetch('auth_type');
my $AUTH			= AbsenceConfig::fetch('authentication');
my $SECRET			= AbsenceConfig::fetch('secret');
my $TOP_PAGE		= AbsenceConfig::fetch('top_page');
my $CRED_SRC		= AbsenceConfig::fetch('credential_src');
my $SESSION_FILE	= AbsenceConfig::fetch('session_file');
my $SESSION_TIMEOUT	= AbsenceConfig::fetch('session_timeout');
my $PW_HASH_FORMAT	= AbsenceConfig::fetch('pw_hash_format');

my $SESSION_FH;

my $DEBUG		= 0;
my $VERSION		= '1.0';
my $AUTH_USER;

#---------------------------------------------------------------------
# class methods
#---------------------------------------------------------------------

sub checkAuthentication
{
	my $q = shift;

	%cookie = $q->cookie($COOKIE_NAME);

	$AUTH || return undef;

	if ($AUTH_TYPE eq 'http') {
		if (exists($ENV{REMOTE_USER})) {
			$AUTH_USER = $ENV{REMOTE_USER};
			return $ENV{REMOTE_USER};
		}
		notAuthenticated($q);
	} elsif ($AUTH_TYPE eq 'simple') {
		openSessionFile();
		if (%cookie) {
			my ($user, $hash) = ($cookie{user}, $cookie{hash});
			my $cs = checkSession($user, $ENV{REMOTE_ADDR});
			if (($hash eq md5_base64($user, $SECRET)) && ($cs eq 'ok')) {
				closeSessionFile();
				$AUTH_USER = $user;
				return $user;
			} else {
				_deleteSession($user, $ENV{REMOTE_ADDR});
			}
		}
		notAuthenticated($q);
		closeSessionFile();
	}

	exit;
}

sub authUser
{
	@_	? $AUTH_USER = shift
		: $AUTH_USER;
}

sub logout
{
	my ($q, $user) = @_;


	if ($AUTH_TYPE eq 'simple') {
		deleteSession($user, $ENV{REMOTE_ADDR});
		my $cook = $q->cookie(
			-name		=> $COOKIE_NAME,
			-value		=> {
				user		=> $user,
				hash		=> md5_base64($Q::user, $SECRET),
			},
			-expires	=> '-10y',
			-path		=> '/',
			-domain		=> AbsenceConfig::fetch('cookie_domain'),
		);
		print $q->header(
				-cookie	=> $cook,
			),
			$q->start_html(
				-title		=> 'Absence: logged off',
				-BGCOLOR	=> '#acacac',
			),
			"logged off.",
			$q->end_html();
	} elsif ($AUTH_TYPE eq 'http') {
		print $q->header(
			-type				=> 'text/html',
			-status				=> '401 Authorization required',
			-nph				=> 1,
			-www_authenticate	=> 'Basic realm="Password Required"',
			),
			$q->start_html(
				-title	=> 'Absence: logged off',
				-BGCOLOR => '#acacac',
			),
			"logged off.",
			$q->end_html();
	}
}

sub authenticateUser
{
	my $q = shift;

	%cookie = $q->cookie($COOKIE_NAME);

	#dbg("authenticateUser: AUTH=[$AUTH]");

	$AUTH || return undef;

	if ($AUTH_TYPE eq 'http') {
		if (exists($ENV{REMOTE_USER})) {
			$AUTH_USER = $ENV{REMOTE_USER};
			return $ENV{REMOTE_USER};
		}
		notAuthenticated($q);
	} elsif ($AUTH_TYPE eq 'simple') {
		openSessionFile();
		if (length($Q::action) && ($Q::action eq 'Login')) {
			if (checkPassword($Q::user, $Q::password)) {
				my $cook = $q->cookie(
					-name	=> $COOKIE_NAME,
					-value	=> {
						user	=> $Q::user,
						hash	=> md5_base64($Q::user, $SECRET),
					},
					-path	=> '/',
					-domain	=> AbsenceConfig::fetch('cookie_domain'),
				);
				dbg("add session: user=[$Q::user], addr=[$ENV{REMOTE_ADDR}]");
				_addSession($Q::user, $ENV{REMOTE_ADDR});
				reloadTop($q, $cook);
			} else {
				authenticationFailed($q);
			}
		} elsif (%cookie) {
			my ($user, $hash) = ($cookie{user}, $cookie{hash});
			my $cs = checkSession($user, $ENV{REMOTE_ADDR});
			my $msg;
			if ($hash eq md5_base64($user, $SECRET)) {
				if ($cs eq 'ok') {
					closeSessionFile();
					$AUTH_USER = $user;
					return $user;
				} elsif ($cs eq 'timeout') {
					$msg = 'Login timed out';
					dbg("authUser: timeout");
				}
			}
			_deleteSession($user, $ENV{REMOTE_ADDR});
			authenticationForm($q, $msg);
		} else {
			authenticationForm($q);
		}
		closeSessionFile();
	}

	exit;
}

sub reloadTop
{
	my ($q, $cook) = @_;
	print $q->header(
			-cookie	=> $cook,
		),
		$q->start_html(
			-title	=> 'Absence: Authentication OK',
			-BGCOLOR => '#acacac',
			-script	=> "top.location = '$TOP_PAGE';",
		),
		$q->start_form(
			-action		=> $TOP_PAGE,
			-method		=> 'POST',
			-target		=> '_parent',
		),
		$q->submit(action => 'Authentication OK. Click to Continue...'),
		$q->end_form(),
		$q->end_html();
}

sub checkPassword
{
	my ($user, $pass) = @_;

	if ($CRED_SRC eq 'absence') {
		#-----------------------------------------------------------
		# simple built-in authentication
		#-----------------------------------------------------------
		my $uref = AbsenceDB::getUser($user);
		defined($uref) || return 0;
		my $pw = ($PW_HASH_FORMAT eq 'md5')
			? md5_base64($pass)
			: $pass;
		($pw eq $uref->{password}) && return 1;
	}

	return 0;
}

sub authenticationForm
{
	my ($q, $msg) = @_;

	print $q->header,
		$q->start_html(
			-title	=> 'Absence Authentication',
			-BGCOLOR => '#acacac',
		);
	if ($msg) { print "<H2>$msg</H2>\n<P>\n"; }
	print "<H2>Please log in</H2>\n",
		$q->start_form(-method => 'POST'),
		'<TABLE><TR><TH ALIGN="right">USER:</TH><TD>',
		$q->textfield(
			-name		=> 'user',
			-size		=> 16,
			-maxlength	=> 16,
		),
		qq[</TD></TR>\n<TR><TH ALIGN="right">PASSWORD:</TH><TD>],
		$q->password_field(
			-name		=> 'password',
			-size		=> 16,
			-maxlength	=> 16,
		),
		"</TD></TR></TABLE><P>",
		$q->submit(action => 'Login'),
		$q->end_form(),
		$q->end_html();
}

sub authenticationFailed
{
	my $q = shift;

	print $q->header(-expires => 'now'),
		$q->start_html('Absence authentication failed'),
		$q->h1('Authentication Failed'),
		$q->end_html;
}

sub notAuthenticated
{
	my $q = shift;

	print $q->header(-expires => 'now'),
		$q->start_html(
			-title		=> 'Absence: not authenticated',
			-BGCOLOR	=> '#ff9081',
		),
		$q->h1('Not Authenticated'),
		$q->end_html;
}

#---------------------------------------------------------------------
# session routines
#---------------------------------------------------------------------

sub checkSession
{
	my ($user, $addr) = @_;

	dbg("checkSession: user=[$user], addr=[$addr]");

	exists($SESSIONS{$user}) && exists($SESSIONS{$user}->{$addr}) || return undef;

	if ($SESSION_TIMEOUT) {
		if ((time() - $SESSIONS{$user}->{$addr}) > $SESSION_TIMEOUT * 60) {
			#deleteSession($user, $addr);
			dbg('session timed out');
			return 'timeout';
		} else {
			# update counter
			_addSession($user, $addr);
		}
	}

	return 'ok';
}

sub _deleteSession
{
	my ($user, $addr) = @_;

	exists($SESSIONS{$user}) && exists($SESSIONS{$user}->{$addr}) || return 0;

	dbg("deleteSession: user=[$user], addr=[$addr]");
	delete($SESSIONS{$user}->{$addr});
	$SESSINFO_CHANGED = 1;
}

sub deleteSession
{
	my ($user, $addr) = @_;

	openSessionFile();

	_deleteSession($user, $addr);

	closeSessionFile();
}

#---------------------------------------------
# add a new session to database
#---------------------------------------------
sub _addSession
{
	my ($user, $addr) = @_;

	dbg("_addSession: user=[$user], addr=[$addr]");
	$SESSIONS{$user}->{$addr} = time();

	$SESSINFO_CHANGED = 1;
}

sub addSession
{
	my ($user, $addr) = @_;

	openSessionFile();

	_addSession($user, $addr);

	closeSessionFile();
}

#---------------------------------------------
# read without locking
#---------------------------------------------
sub readSessionFile
{
	openSessionFile();
	_readSessionFile();
	closeSessionFile();
}

#---------------------------------------------
# expects $SESSION_FH already open
#---------------------------------------------
sub _readSessionFile
{
	dbg("_readSessionFile: top");
	while(<$SESSION_FH>) {
		if (/^((?:\d+\.){3}\d+)\s+(\S+)\s+(\d+)$/) {
			my ($ip_addr, $user, $ts) = ($1, $2, $3);
			dbg("_readSessionFile: ($user, $ip_addr, $ts)");
			$SESSIONS{$user}->{$ip_addr} = $ts;
		}
	}
}

sub _writeSessionFile
{
	dbg("_writeSessionFile: top");
	seek($SESSION_FH, 0, 0) || confess "seek";
	truncate($SESSION_FH, 0) || croak "truncate";
	foreach my $user (keys(%SESSIONS)) {
		foreach my $addr (keys(%{$SESSIONS{$user}})) {
			dbg("_writeSessionFile: ($user, $addr)");
			print $SESSION_FH "$addr\t$user\t$SESSIONS{$user}->{$addr}\n";
		}
	}
}

sub unlockSessionFile
{
	flock($SESSION_FH, LOCK_UN) || die "flock lock_un on $SESSION_FILE ($!)";
}

sub lockSessionFile
{
	flock($SESSION_FH, LOCK_EX) || die "flock lock_ex on $SESSION_FILE ($!)";
}

sub openSessionFile
{
	my $openstr = (-e $SESSION_FILE) ? "+<$SESSION_FILE" : ">$SESSION_FILE";
	dbg("openSessionFile: openstr=[$openstr]");
	$SESSION_FH = FileHandle->new($openstr) || die "open $SESSION_FILE failed";
	lockSessionFile();
	_readSessionFile();
	$SESSINFO_CHANGED = 0;
}

sub closeSessionFile
{
	$SESSINFO_CHANGED && _writeSessionFile();
	unlockSessionFile();
	$SESSION_FH->close;
}

sub dbg
{
	my $msg = shift;
	$DEBUG || return;
	abslog("auth: $msg");
}

1;
